@@ -13,6 +13,19 @@ AUTHOR:		smartfish kafa
 #include <stdio.h>
 #include <qsocket.h>
 #include <qcstring.h>
+#include <malloc.h>
+#include <stdlib.h>
+#include <ctype.h>
+
+/* hack from wget/http.c */
+#define BASE64_LENGTH(len) (4 * (((len) + 2) / 3))
+static void
+base64_encode (const char *s, char *store, int length);
+static char *
+basic_authentication_encode (const char *user, const char *passwd,
+							     const char *header);
+static int
+parse_http_status_line (const char *line, const char **reason_phrase_ptr);
 
 
 const char QTermTelnet::wingate_enter = CTRL('J');
@@ -44,6 +57,7 @@ struct fsm_trans QTermTelnet::ttstab[] = {
 	{ TSWOPT,	TOTXBINARY,	TSDATA,		&QTermTelnet::do_txbinary	},
 	{ TSWOPT,	TCANY,		TSDATA,		&QTermTelnet::do_notsup	},
 
+//	{ TSDOPT,	TONAWS,		TSDATA,		&QTermTelnet::will_naws	},
 	{ TSDOPT,	TOTERMTYPE,	TSDATA,		&QTermTelnet::will_termtype	},
 	{ TSDOPT,	TOTXBINARY,	TSDATA,		&QTermTelnet::will_txbinary	},
 	{ TSDOPT,	TCANY,		TSDATA,		&QTermTelnet::will_notsup	},
@@ -71,12 +85,16 @@ struct fsm_trans QTermTelnet::substab[] = {
  * Constructor
  *------------------------------------------------------------------------
  */
-QTermTelnet::QTermTelnet()
+QTermTelnet::QTermTelnet( QCString cstrTermType )
 {
 	term = new char[21];
 	int i;
 	for(i=0;i<21;i++) term[i]='\000';	//clean up, need???
-	sprintf(term, "vt100");
+	sprintf(term, cstrTermType);
+
+	wx=80;
+	wy=24;
+	done_naws=0;
 
 	synching = 0;
 	doecho = 0;
@@ -84,7 +102,8 @@ QTermTelnet::QTermTelnet()
 	rcvbinary = 0;
 	noga = 0;
 	termtype = 0;
-
+	naws = 0;
+	
 	// proxy related
 	proxy_type = NOPROXY;
 	proxy_state = 0;
@@ -196,10 +215,9 @@ void QTermTelnet::connectHost(const QString& hostname, Q_UINT16 portnumber)
 	port=portnumber;
 	addr_host.sin_port=htons(portnumber);
 	
-	//connect
-	if( proxy_type==NOPROXY )
+	if( proxy_type == NOPROXY )
 	{
-		socket->connectToHost( host, port );
+		 socket->connectToHost( host, port );
 	}
 	else
 	{
@@ -220,7 +238,7 @@ void QTermTelnet::connectHost(const QString& hostname, Q_UINT16 portnumber)
 			emit TelnetState(TSEGETHOSTBYNAME);
 			return;
 		}
-	//now only ipv4 support
+		//now only ipv4 support
 		memcpy(&addr_host.sin_addr,hostent->h_addr,4);
 
 		socket->connectToHost( proxy_host, proxy_port );
@@ -229,6 +247,27 @@ void QTermTelnet::connectHost(const QString& hostname, Q_UINT16 portnumber)
 	emit TelnetState( TSRESOLVING );
 
 }
+
+void QTermTelnet::setWindowSize( int x, int y )
+{
+	wx=x;
+	wy=y;
+
+	char cmd[10];
+	
+	cmd[0] = (char)TCIAC;
+	cmd[1] = (char)TCSB;
+	cmd[2] = (char)TONAWS;
+	cmd[3] = (char)(short(x)>>8);
+	cmd[4] = (char)(short(x)%0xff);
+	cmd[5] = (char)(short(y)>>8);
+	cmd[6] = (char)(short(y)%0xff);
+	cmd[7] = (char)TCIAC;
+	cmd[8] = (char)TCSE;
+
+	write(cmd, 9);
+}
+
 /*------------------------------------------------------------------------
  * set proxy
  *-----------------------------------------------------------------------
@@ -268,6 +307,10 @@ void QTermTelnet::connected()
 	QCString strPort;
 	char command[9];
 	memset(command,0,9);
+
+	char *proxyauth;
+	char *request;
+	int len=0;
 	
 	switch( proxy_type )
 	{
@@ -310,6 +353,28 @@ void QTermTelnet::connected()
 		proxy_state=1;
 		emit TelnetState( TSPROXYCONNECTED );
 		return;
+	case HTTP:
+		proxyauth=NULL;
+		if(bauth)
+			proxyauth = basic_authentication_encode
+				(proxy_usr, proxy_pwd, "Proxy-Authorization");
+
+		len = proxyauth!=NULL?strlen(proxyauth):0;
+	
+		request = new char[host.length()+len+81];
+		
+		sprintf(request,
+					"CONNECT %s:%u HTTP/1.0\n"
+					"%s\n", 
+					host.latin1(),port,
+					proxyauth!=NULL?proxyauth:"");
+
+		write(request, strlen(request));
+		delete request;
+		free(proxyauth);
+		proxy_state=1;
+		emit TelnetState( TSPROXYCONNECTED );
+		return;
 	default:
 		emit TelnetState( TSHOSTCONNECTED );
 		return;		
@@ -405,7 +470,7 @@ void QTermTelnet::socketReadyRead()
 			emit TelnetState( TSPROXYERROR );
 			return;
 		}
-		
+
 		if( from_socket->at(1)==90 )
 			emit TelnetState( TSHOSTCONNECTED );
 		else
@@ -414,12 +479,27 @@ void QTermTelnet::socketReadyRead()
 		return;
 	}
 
+	// HTTP Proxy Reply
+	if( proxy_type==HTTP && proxy_state!=0 )
+	{
+		const char *error;
+		int statcode = parse_http_status_line(from_socket->data(), &error);
+		if( statcode>= 200 && statcode<300 )
+			emit TelnetState( TSHOSTCONNECTED );
+		else
+			emit TelnetState( TSPROXYFAIL );
+		proxy_state=0;
+		return;
+	}
+
 	// Socks5 Proxy Reply
 	if( proxy_type==SOCKS5 && proxy_state!=0 )
 	{
 		socks5_reply(nread);
 		return;
 	}
+
+
 	//resize output buffer
 	to_ansi->resize(2*nread);
 	to_socket->resize(4*nread);
@@ -747,6 +827,31 @@ int QTermTelnet::will_termtype(int c)
 	return 0;
 }
 
+int QTermTelnet::will_naws(int c)
+{
+	done_naws = 1;
+
+	if (naws) {
+		if (option_cmd == TCDO)
+			return 0;
+	} else if (option_cmd == TCDONT)
+		return 0;
+
+	naws = !naws;
+
+	if (naws)
+	{
+		if(!done_naws)
+		{
+			putc_down(TCIAC);
+			putc_down(TCWILL);
+			putc_down((char)c);
+		}
+		setWindowSize( wx, wy );
+	}
+	
+	return 0;
+}
 /*------------------------------------------------------------------------
  * subopt - do option subnegotiation FSM transitions
  *------------------------------------------------------------------------
@@ -1022,3 +1127,130 @@ void QTermTelnet::socks5_reply( int nread )
 		return;
 	}
 }
+
+/* hack from wget/http.c */
+/* How many bytes it will take to store LEN bytes in base64.  */
+#define BASE64_LENGTH(len) (4 * (((len) + 2) / 3))
+
+/* Encode the string S of length LENGTH to base64 format and place it
+   to STORE.  STORE will be 0-terminated, and must point to a writable
+   buffer of at least 1+BASE64_LENGTH(length) bytes.  */
+static void
+base64_encode (const char *s, char *store, int length)
+{
+  /* Conversion table.  */
+  static char tbl[64] = {
+    'A','B','C','D','E','F','G','H',
+    'I','J','K','L','M','N','O','P',
+    'Q','R','S','T','U','V','W','X',
+    'Y','Z','a','b','c','d','e','f',
+    'g','h','i','j','k','l','m','n',
+    'o','p','q','r','s','t','u','v',
+    'w','x','y','z','0','1','2','3',
+    '4','5','6','7','8','9','+','/'
+  };
+  int i;
+  unsigned char *p = (unsigned char *)store;
+
+  /* Transform the 3x8 bits to 4x6 bits, as required by base64.  */
+  for (i = 0; i < length; i += 3)
+    {
+      *p++ = tbl[s[0] >> 2];
+      *p++ = tbl[((s[0] & 3) << 4) + (s[1] >> 4)];
+      *p++ = tbl[((s[1] & 0xf) << 2) + (s[2] >> 6)];
+      *p++ = tbl[s[2] & 0x3f];
+      s += 3;
+    }
+  /* Pad the result if necessary...  */
+  if (i == length + 1)
+    *(p - 1) = '=';
+  else if (i == length + 2)
+    *(p - 1) = *(p - 2) = '=';
+  /* ...and zero-terminate it.  */
+  *p = '\0';
+}
+
+/* Create the authentication header contents for the `Basic' scheme.
+   This is done by encoding the string `USER:PASS' in base64 and
+   prepending `HEADER: Basic ' to it.  */
+static char *
+basic_authentication_encode (const char *user, const char *passwd,
+			     const char *header)
+{
+  char *t1, *t2, *res;
+  int len1 = strlen (user) + 1 + strlen (passwd);
+  int len2 = BASE64_LENGTH (len1);
+
+  t1 = (char *)alloca (len1 + 1);
+  sprintf (t1, "%s:%s", user, passwd);
+  t2 = (char *)alloca (1 + len2);
+  base64_encode (t1, t2, len1);
+  res = (char *)malloc (len2 + 11 + strlen (header));
+  sprintf (res, "%s: Basic %s\r\n", header, t2);
+
+  return res;
+}
+
+/* Parse the HTTP status line, which is of format:
+
+   HTTP-Version SP Status-Code SP Reason-Phrase
+
+   The function returns the status-code, or -1 if the status line is
+   malformed.  The pointer to reason-phrase is returned in RP.  */
+static int
+parse_http_status_line (const char *line, const char **reason_phrase_ptr)
+{
+  /* (the variables must not be named `major' and `minor', because
+     that breaks compilation with SunOS4 cc.)  */
+  int mjr, mnr, statcode;
+  const char *p;
+
+  *reason_phrase_ptr = NULL;
+
+  /* The standard format of HTTP-Version is: `HTTP/X.Y', where X is
+     major version, and Y is minor version.  */
+  if (strncmp (line, "HTTP/", 5) != 0)
+    return -1;
+  line += 5;
+
+  /* Calculate major HTTP version.  */
+  p = line;
+  for (mjr = 0; isdigit (*line); line++)
+    mjr = 10 * mjr + (*line - '0');
+  if (*line != '.' || p == line)
+    return -1;
+  ++line;
+
+  /* Calculate minor HTTP version.  */
+  p = line;
+  for (mnr = 0; isdigit (*line); line++)
+    mnr = 10 * mnr + (*line - '0');
+  if (*line != ' ' || p == line)
+    return -1;
+  /* Wget will accept only 1.0 and higher HTTP-versions.  The value of
+     minor version can be safely ignored.  */
+  if (mjr < 1)
+    return -1;
+  ++line;
+
+  /* Calculate status code.  */
+  if (!(isdigit (*line) && isdigit (line[1]) && isdigit (line[2])))
+    return -1;
+  statcode = 100 * (*line - '0') + 10 * (line[1] - '0') + (line[2] - '0');
+
+  /* Set up the reason phrase pointer.  */
+  line += 3;
+  /* RFC2068 requires SPC here, but we allow the string to finish
+     here, in case no reason-phrase is present.  */
+  if (*line != ' ')
+    {
+      if (!*line)
+	*reason_phrase_ptr = line;
+      else
+	return -1;
+    }
+  else
+    *reason_phrase_ptr = line + 1;
+
+  return statcode;
+}
